extension Array {

	// Safely lookup an index that might be out of bounds, returning nil if it does not exist
	public subscript(safe index: Int) -> Element? {
		get {
			return indices ~= index ? self[index] : nil
		}
		set(newValue) {
			guard let newValue = newValue, indices ~= index else {
				return
			}
			self[index] = newValue
		}
	}

}


// MARK: - Enums

enum SectionInsetType {
	case fixed(sectionInsets: UIEdgeInsets)
	case maxWidth(minimumSectionInsets: UIEdgeInsets, maxWidth: CGFloat)

	func sectionInsets(withCollectionViewWidth collectionViewWidth: CGFloat) -> UIEdgeInsets {
		switch self {
			case .fixed(let sectionInsets):
				return sectionInsets
			case .maxWidth(let minimumSectionInsets, let maxWidth):
				let maxAvailableContentWidth = collectionViewWidth - minimumSectionInsets.left - minimumSectionInsets.right
				let tooWideForMaxWidth = max(0, maxAvailableContentWidth - maxWidth)
				let newLeft = minimumSectionInsets.left + tooWideForMaxWidth / 2
				let newRight = minimumSectionInsets.right + tooWideForMaxWidth / 2
				return UIEdgeInsetsMake(minimumSectionInsets.top, newLeft, minimumSectionInsets.bottom, newRight)
		}
	}
}

// MARK: - Protocols

protocol WHCollectionViewModelInterface {
	func numberOfSections() -> Int
	func numberOfItems(inSection section: Int) -> Int
	func isEmpty() -> Bool
	func canSelectItem(at indexPath: IndexPath) -> Bool
	func dequeueAndConfigureCell(collectionView: UICollectionView, indexPath: IndexPath) -> UICollectionViewCell
	func cellHeight(at indexPath: IndexPath, cellWidth: CGFloat, traitCollection: UITraitCollection) -> CGFloat
	func executePressCellAction(indexPath: IndexPath)
	func headerTitle(inSection section: Int) -> String
	func sectionInsets(inSection section: Int, collectionViewWidth: CGFloat) -> UIEdgeInsets
	func contentWidth(inSection section: Int, collectionViewWidth: CGFloat) -> CGFloat
	func lineSpacing(inSection section: Int) -> CGFloat
}

protocol WHCollectionViewSupplementaryModelInterface {
	mutating func setHeaderHidden(hidden: Bool)
	mutating func setFooterHidden(hidden: Bool)
	func headerSize(inSection section: Int, headerWidth: CGFloat, traitCollection: UITraitCollection) -> CGSize
	func footerSize(inSection section: Int, footerWidth: CGFloat, traitCollection: UITraitCollection) -> CGSize
	func headerInsets(inSection section: Int, collectionViewWidth: CGFloat) -> UIEdgeInsets
	func footerInsets(inSection section: Int, collectionViewWidth: CGFloat) -> UIEdgeInsets
	func dequeueAndConfigureSupplementaryView(withKind kind: String, collectionView: UICollectionView, indexPath: IndexPath) -> UICollectionReusableView
}

// Maybe merge with `PrototypeCellHeightProviding`
protocol PrototypeViewHeightProviding {
	func fittingSizeForConfiguredView(compactLayout: Bool, width: CGFloat) -> CGSize
}

typealias WHCollectionViewAndSupplementaryModelInterface = WHCollectionViewModelInterface & WHCollectionViewSupplementaryModelInterface

// MARK: - Extensions

extension WHCollectionViewModelInterface {

	func isEmpty() -> Bool {
		var items = 0
		for section in 0..<numberOfSections() {
			items += numberOfItems(inSection: section)
		}

		return items == 0
	}

	func contentWidth(inSection section: Int, collectionViewWidth: CGFloat) -> CGFloat {
		let insets = sectionInsets(inSection: section, collectionViewWidth: collectionViewWidth)
		return max(0, collectionViewWidth - insets.left - insets.right)
	}

}

extension WHCollectionViewSupplementaryModelInterface {

	func headerWidth(inSection section: Int, collectionViewWidth: CGFloat) -> CGFloat {
		let insets = headerInsets(inSection: section, collectionViewWidth: collectionViewWidth)
		return max(0, collectionViewWidth - insets.left - insets.right)
	}

	func footerWidth(inSection section: Int, collectionViewWidth: CGFloat) -> CGFloat {
		let insets = footerInsets(inSection: section, collectionViewWidth: collectionViewWidth)
		return max(0, collectionViewWidth - insets.left - insets.right)
	}

}

// MARK: - Structs

struct WHCollectionViewSectionsModel: WHCollectionViewModelInterface, WHCollectionViewSupplementaryModelInterface, CellSelectable {

	private(set) var sectionEntries: [WHCollectionViewAndSupplementaryModelInterface]

	init(sectionEntries: [WHCollectionViewAndSupplementaryModelInterface]) {
		self.sectionEntries = sectionEntries
	}

	// MARK: - WHCollectionViewModelInterface

	func sectionModelAtSectionIndex(_ section: Int) -> WHCollectionViewAndSupplementaryModelInterface? {
		return sectionEntries[safe: section]
	}

	func cellHeight(at indexPath: IndexPath, cellWidth: CGFloat, traitCollection: UITraitCollection) -> CGFloat {
		return sectionModelAtSectionIndex(indexPath.section)?.cellHeight(at: indexPath, cellWidth: cellWidth, traitCollection: traitCollection) ?? 0
	}

	func canSelectItem(at indexPath: IndexPath) -> Bool {
		guard let sectionModel = sectionModelAtSectionIndex(indexPath.section) else { return false }
		return sectionModel.canSelectItem(at: indexPath)
	}

	func executePressCellAction(indexPath: IndexPath) {
		sectionModelAtSectionIndex(indexPath.section)?.executePressCellAction(indexPath: indexPath)
	}

	func numberOfSections() -> Int {
		return sectionEntries.count
	}

	func numberOfItems(inSection section: Int) -> Int {
		return sectionModelAtSectionIndex(section)?.numberOfItems(inSection: section) ?? 0
	}

	func dequeueAndConfigureCell(collectionView: UICollectionView, indexPath: IndexPath) -> UICollectionViewCell {
		guard let sectionM = sectionModelAtSectionIndex(indexPath.section) else { fatalError("Tried to reach for the sectionModel with an out of bounds indexPath.") }
		return sectionM.dequeueAndConfigureCell(collectionView: collectionView, indexPath: indexPath)
	}

	func headerTitle(inSection section: Int) -> String {
		return sectionModelAtSectionIndex(section)?.headerTitle(inSection: section) ?? ""
	}

	func sectionInsets(inSection section: Int, collectionViewWidth: CGFloat) -> UIEdgeInsets {
		return sectionModelAtSectionIndex(section)?.sectionInsets(inSection: section, collectionViewWidth: collectionViewWidth) ?? UIEdgeInsets.zero
	}
	func lineSpacing(inSection section: Int) -> CGFloat {
		return sectionModelAtSectionIndex(section)?.lineSpacing(inSection: section) ?? 0
	}

	// MARK: - WHCollectionViewSupplementaryModelInterface

	mutating func setHeaderHidden(hidden: Bool) {
		sectionEntries = sectionEntries.map { section in
			var section = section
			section.setHeaderHidden(hidden: hidden)
			return section
		}
	}

	mutating func setFooterHidden(hidden: Bool) {
		sectionEntries = sectionEntries.map { section in
			var section = section
			section.setFooterHidden(hidden: hidden)
			return section
		}
	}

	func headerSize(inSection section: Int, headerWidth: CGFloat, traitCollection: UITraitCollection) -> CGSize {
		return sectionModelAtSectionIndex(section)?.headerSize(inSection: section, headerWidth: headerWidth, traitCollection: traitCollection) ?? .zero
	}

	func footerSize(inSection section: Int, footerWidth: CGFloat, traitCollection: UITraitCollection) -> CGSize {
		return sectionModelAtSectionIndex(section)?.footerSize(inSection: section, footerWidth: footerWidth, traitCollection: traitCollection) ?? .zero
	}

	func headerInsets(inSection section: Int, collectionViewWidth: CGFloat) -> UIEdgeInsets {
		return sectionModelAtSectionIndex(section)?.headerInsets(inSection: section, collectionViewWidth: collectionViewWidth) ?? .zero
	}

	func footerInsets(inSection section: Int, collectionViewWidth: CGFloat) -> UIEdgeInsets {
		return sectionModelAtSectionIndex(section)?.footerInsets(inSection: section, collectionViewWidth: collectionViewWidth) ?? .zero
	}

	func dequeueAndConfigureSupplementaryView(withKind kind: String, collectionView: UICollectionView, indexPath: IndexPath) -> UICollectionReusableView {
		guard let sectionM = sectionModelAtSectionIndex(indexPath.section) else { fatalError("Tried to reach for the sectionModel with an out of bounds indexPath.") }
		return sectionM.dequeueAndConfigureSupplementaryView(withKind: kind, collectionView: collectionView, indexPath: indexPath)
	}

	// MARK: - CellSelectable

	mutating func addCellToSelection(indexPath: IndexPath) {
		guard canSelectItem(at: indexPath) else { return }
		sectionEntries = sectionEntriesByMappingOnCellSelectables({ section, cellSelectable in
			guard section == (indexPath).section else { return cellSelectable }
			var newCellSelectable = cellSelectable
			newCellSelectable.addCellToSelection(indexPath: indexPath)
			return newCellSelectable
		})
	}

	mutating func removeCellFromSelection(indexPath: IndexPath) {
		guard canSelectItem(at: indexPath) else { return }
		sectionEntries = sectionEntriesByMappingOnCellSelectables({ section, cellSelectable in
			guard section == indexPath.section else { return cellSelectable }
			var newCellSelectable = cellSelectable
			newCellSelectable.removeCellFromSelection(indexPath: indexPath)
			return newCellSelectable
		})
	}

	mutating func removeAllCellsFromSelection() {
		sectionEntries = sectionEntriesByMappingOnCellSelectables({ section, cellSelectable in
			var newCellSelectable = cellSelectable
			newCellSelectable.removeAllCellsFromSelection()
			return newCellSelectable
		})
	}

	private typealias CellSelectableCollectionViewModelInterface = WHCollectionViewAndSupplementaryModelInterface & CellSelectable

	// needed, as arrays are structs, and mutating elements in an array property does not work otherwise
	private func sectionEntriesByMappingOnCellSelectables(_ f: (_ section: Int, _ cellSelectable: CellSelectableCollectionViewModelInterface) -> CellSelectableCollectionViewModelInterface) -> [WHCollectionViewAndSupplementaryModelInterface] {
		return zip(sectionEntries.indices, sectionEntries).map({ (index: Int, element: WHCollectionViewAndSupplementaryModelInterface) -> WHCollectionViewAndSupplementaryModelInterface in
			guard let cellSelectable = element as? WHCollectionViewAndSupplementaryModelInterface & CellSelectable else { return element }
			return f(index, cellSelectable)
		})
	}

	func isCellSelected(indexPath: IndexPath) -> Bool {
		return cellSelectableSectionEntry(atSectionIndex: indexPath.section)?.isCellSelected(indexPath: indexPath) ?? false
	}

	var cellSelectionCount: Int {
		return cellSelectableSectionEntries().map({ $0.cellSelectionCount }).reduce(0, +)
	}

	func executeActionForCellSelection() {
		cellSelectableSectionEntries().forEach { selectable in
			selectable.executeActionForCellSelection()
		}
	}

	func cellSelectableSectionEntry(atSectionIndex section: Int) -> CellSelectable? {
		return sectionModelAtSectionIndex(section) as? CellSelectable
	}

	func cellSelectableSectionEntries() -> [CellSelectable] {
		return sectionEntries.map({ return $0 as? CellSelectable }).compactMap({ $0 })
	}

}

struct WHCollectionViewSingleSectionModelWithEqualCellHeights<M, C: UICollectionViewCell>: WHCollectionViewModelInterface, WHCollectionViewSupplementaryModelInterface, CellSelectable where M: Hashable, M: SelectableReporting, C: HeightProviding, C: Reusable {

	var cellSelectionCount: Int {
		return selectedCellModels.count
	}

	private(set) var configureCell: (_ cell: C, _ cellModel: M, _ traitCollection: UITraitCollection, _ selected: Bool) -> Void
	private(set) var onPressCell: (_ cellModel: M) -> Void
	private(set) var actionForCellSelection: (_ cellSelection: [M]) -> Void

	private var cellModels: [M]
	private var selectedCellModels = Set<M>()
	private var supplementaryModel: WHCollectionViewSupplementaryModelInterface?

	private let sectionHeaderTitle: String
	private let sectionInsetType: SectionInsetType

	private let lineSpacing: CGFloat

	init(cellModels: [M], configureCell: @escaping (_ cell: C, _ cellModel: M, _ traitCollection: UITraitCollection, _ selected: Bool) -> Void, onPressCell: @escaping (_ cellModel: M) -> Void, actionForCellSelection: @escaping (_ cellSelection: [M]) -> Void = { _ in }, sectionInsetType: SectionInsetType = .fixed(sectionInsets: .zero), headerTitle: String = "", supplementaryModel: WHCollectionViewSupplementaryModelInterface? = nil, lineSpacing: CGFloat = 0) {
		self.cellModels = cellModels
		self.configureCell = configureCell
		self.onPressCell = onPressCell
		self.actionForCellSelection = actionForCellSelection
		self.sectionInsetType = sectionInsetType
		self.sectionHeaderTitle = headerTitle
		self.supplementaryModel = supplementaryModel
		self.lineSpacing = lineSpacing
	}

	func dequeueAndConfigureCell(collectionView: UICollectionView, indexPath: IndexPath) -> UICollectionViewCell {
		let cell: C = collectionView.wh_dequeueReusableCell(forIndexPath: indexPath)

		guard let model = self.cellModel(atIndexPath: indexPath) else {
			// cannot configure
			return cell
		}

		let selected = selectedCellModels.contains(model)
		configureCell(cell, model, collectionView.traitCollection, selected)
		return cell
	}

	func canSelectItem(at indexPath: IndexPath) -> Bool {
		guard let model = cellModel(atIndexPath: indexPath) else { return false }
		return model.canSelectItem()
	}

	func executePressCellAction(indexPath: IndexPath) {
		guard let model = self.cellModel(atIndexPath: indexPath) else {
			// ignore press
			return
		}
		onPressCell(model)
	}

	func cellHeight(at indexPath: IndexPath, cellWidth: CGFloat, traitCollection: UITraitCollection) -> CGFloat {
		return C.height(forWidth: cellWidth, traitCollection: traitCollection)
	}

	func numberOfSections() -> Int {
		return 1
	}

	func numberOfItems(inSection section: Int) -> Int {
		return cellModels.count
	}

	func headerTitle(inSection section: Int) -> String {
		return sectionHeaderTitle
	}

	func sectionInsets(inSection section: Int, collectionViewWidth: CGFloat) -> UIEdgeInsets {
		return sectionInsetType.sectionInsets(withCollectionViewWidth: collectionViewWidth)
	}

	func lineSpacing(inSection section: Int) -> CGFloat {
		return lineSpacing
	}

	mutating func addCellToSelection(indexPath: IndexPath) {
		guard canSelectItem(at: indexPath) else { return }
		guard let model = self.cellModel(atIndexPath: indexPath) else {
			// ignore
			return
		}
		selectedCellModels.insert(model)
	}

	mutating func removeCellFromSelection(indexPath: IndexPath) {
		guard canSelectItem(at: indexPath) else { return }
		guard let model = cellModel(atIndexPath: indexPath) else {
			// ignore
			return
		}
		selectedCellModels.remove(model)
	}

	mutating func removeAllCellsFromSelection() {
		selectedCellModels.removeAll()
	}

	func isCellSelected(indexPath: IndexPath) -> Bool {
		guard let model = self.cellModel(atIndexPath: indexPath) else {
			return false
		}

		return selectedCellModels.contains(model)
	}

	func executeActionForCellSelection() {
		actionForCellSelection(Array(selectedCellModels))
	}

	func cellModel(atIndexPath indexPath: IndexPath) -> M? {
		let model = cellModels[safe: Int(indexPath.row)]
		assert(model != nil)
		return model
	}

	var allCellModels: [M] {
		return cellModels
	}

	func indexPath(forCellModel searchedCellModel: M, section: Int) -> IndexPath? {
		return indexPath(forCellModelSatisfying: { cellModel in
			return searchedCellModel == cellModel
		}, section: section)
	}

	func indexPath(forCellModelSatisfying satisfying: (M) -> Bool, section: Int) -> IndexPath? {
		guard let index = cellModels.index(where: satisfying) else {
			return nil
		}
		return IndexPath(row: index, section: section)
	}

	func visibleCell(inCollectionView collectionView: UICollectionView, forCellModelSatisfying satisfying: (M) -> Bool, section: Int) -> C? {
		guard let indexPath = indexPath(forCellModelSatisfying: satisfying, section: section), let visibleCell = collectionView.cellForItem(at: indexPath) as? C else {
			return nil
		}
		return visibleCell
	}

	mutating func setHeaderHidden(hidden: Bool) {
		supplementaryModel?.setHeaderHidden(hidden: hidden)
	}

	mutating func setFooterHidden(hidden: Bool) {
		supplementaryModel?.setFooterHidden(hidden: hidden)
	}

	func headerSize(inSection section: Int, headerWidth: CGFloat, traitCollection: UITraitCollection) -> CGSize {
		guard !cellModels.isEmpty else { return CGSize.zero }
		guard let supplementaryModel = supplementaryModel else { return CGSize.zero }
		return supplementaryModel.headerSize(inSection: section, headerWidth: headerWidth, traitCollection: traitCollection)
	}

	func footerSize(inSection section: Int, footerWidth: CGFloat, traitCollection: UITraitCollection) -> CGSize {
		guard !cellModels.isEmpty else { return CGSize.zero }
		guard let supplementaryModel = supplementaryModel else { return CGSize.zero }
		return supplementaryModel.footerSize(inSection: section, footerWidth: footerWidth, traitCollection: traitCollection)
	}

	func headerInsets(inSection section: Int, collectionViewWidth: CGFloat) -> UIEdgeInsets {
		return supplementaryModel?.headerInsets(inSection: section, collectionViewWidth: collectionViewWidth) ?? .zero
	}

	func footerInsets(inSection section: Int, collectionViewWidth: CGFloat) -> UIEdgeInsets {
		return supplementaryModel?.footerInsets(inSection: section, collectionViewWidth: collectionViewWidth) ?? .zero
	}

	func dequeueAndConfigureSupplementaryView(withKind kind: String, collectionView: UICollectionView, indexPath: IndexPath) -> UICollectionReusableView {
		guard let view = supplementaryModel?.dequeueAndConfigureSupplementaryView(withKind: kind, collectionView: collectionView, indexPath: indexPath) else {
			assertionFailure()
			return UICollectionReusableView()
		}
		return view
	}

}


class CollectionViewFlowLayoutDataSourceAndDelegate: NSObject {

	private var viewModel = WHCollectionViewSectionsModel(sectionEntries: [])

	func configure(withViewModel viewModel: WHCollectionViewSectionsModel, collectionView: UICollectionView) {
		self.viewModel = viewModel
		collectionView.reloadData()
	}

}

// MARK: - UICollectionViewDataSource

extension CollectionViewFlowLayoutDataSourceAndDelegate: UICollectionViewDataSource {

	func numberOfSections(in collectionView: UICollectionView) -> Int {
		return viewModel.numberOfSections()
	}

	func collectionView(_ collectionView: UICollectionView, numberOfItemsInSection section: Int) -> Int {
		return viewModel.numberOfItems(inSection: section)
	}

	func collectionView(_ collectionView: UICollectionView, cellForItemAt indexPath: IndexPath) -> UICollectionViewCell {
		let cell = viewModel.dequeueAndConfigureCell(collectionView: collectionView, indexPath: indexPath)
		return cell
	}

	func collectionView(_ collectionView: UICollectionView, viewForSupplementaryElementOfKind kind: String, at indexPath: IndexPath) -> UICollectionReusableView {
		return viewModel.dequeueAndConfigureSupplementaryView(withKind: kind, collectionView: collectionView, indexPath: indexPath)
	}
}

// MARK: - ResizeInvalidatingSupplementaryViewInsetCollectionViewFlowLayoutDelegate

extension CollectionViewFlowLayoutDataSourceAndDelegate: ResizeInvalidatingSupplementaryViewInsetCollectionViewFlowLayoutDelegate {

	func collectionView(_ collectionView: UICollectionView, layout collectionViewLayout: UICollectionViewLayout, sizeForItemAt indexPath: IndexPath) -> CGSize {
		let width = viewModel.contentWidth(inSection: indexPath.section, collectionViewWidth: collectionView.bounds.width)
		let height = viewModel.cellHeight(at: indexPath, cellWidth: width, traitCollection: collectionView.traitCollection)
		return CGSize(width: width, height: height)
	}

	func collectionView(_ collectionView: UICollectionView, layout collectionViewLayout: UICollectionViewLayout, referenceSizeForHeaderInSection section: Int) -> CGSize {
		let width = viewModel.headerWidth(inSection: section, collectionViewWidth: collectionView.bounds.width)
		return viewModel.headerSize(inSection: section, headerWidth: width, traitCollection: collectionView.traitCollection)
	}

	func collectionView(_ collectionView: UICollectionView, layout collectionViewLayout: UICollectionViewLayout, referenceSizeForFooterInSection section: Int) -> CGSize {
		let width = viewModel.footerWidth(inSection: section, collectionViewWidth: collectionView.bounds.width)
		return viewModel.footerSize(inSection: section, footerWidth: width, traitCollection: collectionView.traitCollection)
	}

	func collectionView(_ collectionView: UICollectionView, didSelectItemAt indexPath: IndexPath) {
		viewModel.executePressCellAction(indexPath: indexPath)
	}

	func collectionView(_ collectionView: UICollectionView, didDeselectItemAt indexPath: IndexPath) {
		collectionView.configureHighlightableCellForHighlightedOrSelectedState(atIndexPath: indexPath, selectedOrHighlighted: false)
	}

	func collectionView(_ collectionView: UICollectionView, didHighlightItemAt indexPath: IndexPath) {
		collectionView.configureHighlightableCellForHighlightedOrSelectedState(atIndexPath: indexPath, selectedOrHighlighted: true)
	}

	func collectionView(_ collectionView: UICollectionView, shouldHighlightItemAt indexPath: IndexPath) -> Bool {
		return viewModel.canSelectItem(at: indexPath)
	}

	func collectionView(_ collectionView: UICollectionView, shouldSelectItemAt indexPath: IndexPath) -> Bool {
		return viewModel.canSelectItem(at: indexPath)
	}

	func collectionView(_ collectionView: UICollectionView, didUnhighlightItemAt indexPath: IndexPath) {
		collectionView.configureHighlightableCellForHighlightedOrSelectedState(atIndexPath: indexPath, selectedOrHighlighted: false)
	}

	func collectionView(_ collectionView: UICollectionView, layout collectionViewLayout: UICollectionViewLayout, insetForSectionAt section: Int) -> UIEdgeInsets {
		return viewModel.sectionInsets(inSection: section, collectionViewWidth: collectionView.bounds.size.width)
	}

	func collectionView(_ collectionView: UICollectionView, layout collectionViewLayout: ResizeInvalidatingSupplementaryViewInsetCollectionViewFlowLayout, insetForHeaderAt section: Int) -> UIEdgeInsets {
		return viewModel.headerInsets(inSection: section, collectionViewWidth: collectionView.bounds.size.width)
	}

	func collectionView(_ collectionView: UICollectionView, layout collectionViewLayout: ResizeInvalidatingSupplementaryViewInsetCollectionViewFlowLayout, insetForFooterAt section: Int) -> UIEdgeInsets {
		return viewModel.footerInsets(inSection: section, collectionViewWidth: collectionView.bounds.size.width)
	}

}
